@@ -1,23 +1,25 @@
 """ Curses terminal digital clock"""
 import argparse
 import curses
-import time
 from datetime import datetime
 
+from typing import Generator
 from typing import Optional
 from typing import Sequence
 from typing import Tuple
 
 import time_machine
 
-# might not need color black
 CURSES_COLORS = {"black": curses.COLOR_BLACK, "white": curses.COLOR_WHITE,
                  "red": curses.COLOR_RED, "green": curses.COLOR_GREEN,
                  "blue": curses.COLOR_BLUE, "magenta": curses.COLOR_MAGENTA,
                  "yellow": curses.COLOR_YELLOW, "cyan": curses.COLOR_CYAN}
 CHAR_CODES_COLOR = {114: "red", 116: "green", 121: "blue", 117: "yellow",
-                    105: "magenta", 111: "cyan", 112: "white"}
-COLORS = ["red", "green", "blue", "yellow", "magenta", "cyan", "white"]
+                    105: "magenta", 111: "cyan", 112: "white", 91: "black"}
+CHAR_CODES_COLOR_BG = {82: "red", 84: "green", 89: "blue", 85: "yellow",
+                       73: "magenta", 79: "cyan", 80: "white", 123: "black"}
+COLORS = ["red", "green", "blue", "yellow", "magenta", "cyan", "white", "black"]
+DATE_FORMATS = ["%d/%m/%Y", "%m/%d/%Y", "%Y/%m/%d", "%Y/%d/%m"]
 
 
 class CTClockError(Exception):
@@ -79,26 +81,53 @@ class MyTime:
         self.test_mode = test_mode
         self.test_time = test_time
         self.tick = tick
+        self.save_time = "00:00:00"
+        self.paused = False
         if test_mode:
             self.time = self._time_generator()
         else:
             self.time = None
 
-    def _time_generator(self):
+    def _time_generator(self) -> Generator:
         with time_machine.travel(self.test_time, tick=self.tick):
             while True:
                 yield datetime.today()
 
-    def get_time(self, time_format: str):
+    def get_time(self, time_format: str) -> str:
         if self.test_mode:
+            if self.paused:
+                return "".join(self.save_time[0:8].split(":"))
             return next(self.time).strftime(time_format)
         else:
             return datetime.today().strftime(time_format)
 
-    def reset_time(self):
+    def get_date(self, date_format: str) -> str:
+        if self.test_mode:
+            return next(self.time).date().strftime(date_format)
+        else:
+            return datetime.today().date().strftime(date_format)
+
+    def reset_time(self) -> None:
+        # used in the context manager in the test suite
         if self.test_mode:
             self.time.close()
 
+    def pause(self):
+        self.save_time = next(self.time).time().strftime("%H:%M:%S.%f")
+        self.time.close()
+        self.paused = True
+
+    def unpause(self):
+        self.test_time = self.save_time
+        self.time = self._time_generator()
+        self.paused = False
+
+    def reset(self):
+        self.time.close()
+        self.save_time = self.test_time = "00:00:00"
+        self.time = self._time_generator()
+        self.paused = True
+
 
 def get_segments(number: str, size: str) -> tuple:
     if size == "small":
@@ -157,24 +186,37 @@ def get_space_size(size: str, show_seconds: bool) -> Tuple[int, int]:
     return h, w
 
 
-def set_color(color: str) -> None:
+def fill_background(screen, bg_color: str) -> None:
+    curses.init_pair(3, CURSES_COLORS[bg_color], CURSES_COLORS[bg_color])
+    height, width = screen.getmaxyx()
+    for y in range(height - 1):
+        for x in range(width):
+            screen.addstr(y, x, " ", curses.color_pair(3))
+
+
+def set_color(color: str, bg_color: str) -> None:
     curses.init_pair(1, CURSES_COLORS[color], CURSES_COLORS[color])
-    curses.init_pair(2, CURSES_COLORS[color], curses.COLOR_BLACK)
+    curses.init_pair(2, CURSES_COLORS[color], CURSES_COLORS[bg_color])
 
 
 def display(screen, time_string: str, size: str,
             size_x: int, size_y: int, color: str,
             show_seconds: bool, am_pm: str, show_date: bool,
-            colon_on: bool, test_mode: bool, military_time: bool) -> None:
+            colon_on: bool, test_mode: bool, military_time: bool, date: str,
+            bg_color: str, stop_watch: bool, stop_watch_state: str) -> None:
     time_segments = []
     for digit in time_string:
         time_segments.append(get_segments(digit, size))
-    set_color(color)
+    set_color(color, bg_color)
     size_offset = get_offset(size)
     height, width = get_space_size(size, show_seconds)
     hc = int((size_y - height) / 2)  # height/vertical center
     w_offset = int((size_x - width) / 2)  # width/horizontal center
     screen.clear()
+    fill_background(screen, bg_color)
+    if stop_watch:
+        msg = f"Stop Watch  {stop_watch_state}"
+        screen.addstr(hc - 2, int(size_x / 2) - 10, msg, curses.color_pair(2))
     if military_time or time_string[:1] == "1":
         d = "1" if not test_mode else time_string[:1]
         for seg in time_segments[0]:
@@ -212,22 +254,94 @@ def display(screen, time_string: str, size: str,
     if am_pm != "":
         screen.addstr(height + hc, w_offset, am_pm, curses.color_pair(2))
     if show_date:
-        date = datetime.today().date().strftime("%d/%m/%Y")
         screen.addstr(height + hc, w_offset - 15, date, curses.color_pair(2))
     if test_mode:
         screen.addstr(0, 0, "test mode", curses.color_pair(2))
         screen.addstr(1, 0, color, curses.color_pair(2))
+        screen.addstr(2, 0, f"bg={bg_color}", curses.color_pair(2))
     screen.refresh()
 
 
-def main_clock(screen, static_color: str, show_seconds: bool,
-               military_time: bool, screen_saver_mode: bool,
-               mode: int, cycle_timing: int, show_date: bool,
-               blink_colon: bool, test_mode: bool, test_time: str,
-               no_colon: bool) -> None:
+def main_stopwatch(screen, args: argparse.Namespace) -> None:
     curses.curs_set(0)  # Set the cursor to off.
     screen.timeout(0)  # Turn blocking off for screen.getch()
-    ct_time = MyTime(test_mode, test_time, True)
+    bg_color = args.bg_color
+    digit_color = args.color
+    if args.auto_start:
+        state = "Running"
+        paused = False
+    else:
+        state = "Stopped"  # running, stopped, paused
+        paused = True
+    size_y, size_x = screen.getmaxyx()
+    if size_x >= 90 and size_y >= 22:
+        text_size = "large"
+    elif size_x >= 46 and size_y >= 10:
+        text_size = "medium"
+    elif size_x >= 36 and size_y >= 9:
+        text_size = "small"
+    else:
+        raise CTClockError("Error screen / window is to small")
+    update_screen = True
+    ct_clock = MyTime(True, "00:00:00", True)
+    display_time = ct_clock.get_time("%H%M%S")
+    while True:
+        if curses.is_term_resized(size_y, size_x):
+            size_y, size_x = screen.getmaxyx()
+            if size_x >= 90 and size_y >= 22:
+                text_size = "large"
+            elif size_x >= 46 and size_y >= 12:
+                text_size = "medium"
+            elif size_x >= 36 and size_y >= 9:
+                text_size = "small"
+            else:
+                raise CTClockError("Error screen / window is to small")
+            display(screen, display_time, text_size, size_x, size_y, digit_color, True,
+                    "", False, True, args.test_mode, True, "", bg_color, True, state)
+        if update_screen or not paused and display_time != ct_clock.get_time("%H%M%S"):
+            display_time = ct_clock.get_time("%H%M%S")
+            display(screen, display_time, text_size, size_x, size_y, digit_color, True,
+                    "", False, True, args.test_mode, True, "", bg_color, True, state)
+            update_screen = False
+        ch = screen.getch()
+        if ch in [81, 113]:  # q, Q
+            break
+        elif ch == 103:  # g
+            if paused:
+                ct_clock.unpause()
+                paused = False
+                state = "Running"
+                update_screen = True
+            else:
+                ct_clock.pause()
+                paused = True
+                state = "Paused"
+                update_screen = True
+        elif ch == 104:  # h
+            update_screen = True
+            paused = True
+            state = "Stopped"
+            display_time = ct_clock.get_time("%H%M%S")
+            ct_clock.reset()
+        if ch in CHAR_CODES_COLOR.keys():
+            digit_color = CHAR_CODES_COLOR[ch]
+            display(screen, display_time, text_size, size_x, size_y, digit_color, True,
+                    "", False, True, args.test_mode, True, "", bg_color, True, state)
+
+
+def main_clock(screen, args: argparse.Namespace) -> None:
+    static_color = args.color
+    show_seconds = args.no_seconds
+    military_time = args.military_time
+    mode = args.mode
+    cycle_timing = args.cycle_timing
+    show_date = args.show_date
+    blink_colon = args.blink_colon
+    no_colon = args.no_colon
+    bg_color = args.bg_color
+    curses.curs_set(0)  # Set the cursor to off.
+    screen.timeout(0)  # Turn blocking off for screen.getch()
+    ct_time = MyTime(args.test_mode, args.test_date + " " + args.test_time, True)
     update_screen = True
     size_y, size_x = screen.getmaxyx()
     if size_x >= 90 and size_y >= 20:
@@ -240,6 +354,7 @@ def main_clock(screen, static_color: str, show_seconds: bool,
         raise CTClockError("Error screen / window is to small")
 
     time_format = "%H%M%S" if military_time else "%I%M%S"
+    date_format_pointer = 0
 
     colon_on = False if no_colon else True
     cycle_count = 0
@@ -286,18 +401,22 @@ def main_clock(screen, static_color: str, show_seconds: bool,
                 color = COLORS[cycle_count]
             else:
                 color = static_color
+            date = ct_time.get_date(DATE_FORMATS[date_format_pointer])
             display(screen, displayed, text_size, size_x, size_y,
                     color, show_seconds, am_pm, show_date, colon_on,
-                    test_mode, military_time)
+                    args.test_mode, military_time, date, bg_color, False, "")
             update_screen = False
         ch = screen.getch()
-        if screen_saver_mode and ch != -1:
+        if args.screensaver and ch != -1:
             break
         if ch in [81, 113]:  # q, Q
             break
         if mode == 0 and ch in CHAR_CODES_COLOR.keys():
             static_color = CHAR_CODES_COLOR[ch]
             update_screen = True
+        if ch in CHAR_CODES_COLOR_BG.keys():
+            bg_color = CHAR_CODES_COLOR_BG[ch]
+            update_screen = True
         if ch == 99:  # c
             if mode == 1:
                 mode = 0
@@ -324,15 +443,23 @@ def main_clock(screen, static_color: str, show_seconds: bool,
         elif ch == 101:  # e
             show_date = not show_date  # flips between True and False
             update_screen = True
+        elif ch == 69 and show_date:  # E
+            if date_format_pointer == len(DATE_FORMATS) - 1:
+                date_format_pointer = 0
+            else:
+                date_format_pointer += 1
+            update_screen = True
         elif ch == 100:  # d
             time_format = "%I%M%S"
             military_time = False
             show_date = False
+            date_format_pointer = 0
             show_seconds = True
             mode = 0
             static_color = "white"
             blink_colon = False
             colon_on = True
+            bg_color = "black"
             update_screen = True
         elif ch == 110:  # n
             colon_on = not colon_on
@@ -355,12 +482,24 @@ def display_running_commands() -> None:
     print(" c       Change color mode")
     print(" s       Toggle show seconds")
     print(" e       Toggle show date")
+    print(" E       Cycle date formats")
     print(" b       Toggle blink colon")
     print(" m       Toggle military time")
     print(" n       Toggle colon off and on")
     print(" d       Reset setting to defaults")
     print(" 1,2,3   Color cycle timing 1-every second, 2-every minute, 3-every hour")
-    print(" r,t,y,u,i,o,p")
+    print(" r,t,y,u,i,o,p,[")
+    print("         Select color: Red, Green, Blue, Yellow, Magenta, Cyan, White")
+    print(" R,T,Y,U,I,O,P,{")
+    print("         Change background colors: Red, Green, Blue, Yellow, Magenta, "
+          "Cyan, White, Black")
+    print()
+    print()
+    print("STOP WATCH Commands:")
+    print(" q  Q    Quit")
+    print(" g       Start or pause stop watch")
+    print(" h       Reset stop watch to 00:00:00")
+    print(" r,t,y,u,i,o,p,[")
     print("         Select color: Red, Green, Blue, Yellow, Magenta, Cyan, White")
 
 
@@ -395,11 +534,24 @@ def argument_parser(argv: Optional[Sequence[str]] = None) -> argparse.Namespace:
                         help="Cycle timing (1 every sec, 2 every min, 3 every hour)")
     parser.add_argument("--show_date", action="store_true",
                         help="Show date")
+    parser.add_argument("--bg_color", type=color_type, default="black",
+                        help="Back ground color")
     parser.add_argument("--list_commands", action="store_true",
                         help="List commands available during run time.")
     parser.add_argument("--test_mode", action="store_true", help=argparse.SUPPRESS)
     parser.add_argument("--test_time", type=str, default="00:00:00",
                         help=argparse.SUPPRESS)
+    parser.add_argument("--test_date", type=str, default="1970-1-2",
+                        help=argparse.SUPPRESS)
+
+    sub_parser = parser.add_subparsers(dest="command")
+    stop_watch_parser = sub_parser.add_parser("stop_watch")
+    stop_watch_parser.add_argument("--auto_start", action="store_true",
+                                   help="Auto start stop watch")
+    stop_watch_parser.add_argument("-c", "--color", type=color_type, default="white",
+                                   help="digit color")
+    stop_watch_parser.add_argument("--list_commands", action="store_true",
+                                   help="List commands available during run time.")
     return parser.parse_args(argv)
 
 
@@ -408,12 +560,15 @@ def main(argv: Optional[Sequence[str]] = None) -> int:
     if args.list_commands:
         display_running_commands()
         return 0
+    elif args.command == "stop_watch":
+        try:
+            curses.wrapper(main_stopwatch, args)
+        except CTClockError as e:
+            print(e)
+            return 1
+        return 0
     try:
-        curses.wrapper(main_clock, args.color, args.no_seconds,
-                       args.military_time, args.screensaver,
-                       args.mode, args.cycle_timing, args.show_date,
-                       args.blink_colon, args.test_mode, args.test_time,
-                       args.no_colon)
+        curses.wrapper(main_clock, args)
     except CTClockError as e:
         print(e)
         return 1
